// SPDX-License-Identifier: GPL-2.0-only
// Copyright (c) 2018-2020, Intel Corporation
//
// sof-es8336.c - ASoC machine driver for Up and Up2 board
// based on WM8804/Hifiberry Digi+


#include <linux/acpi.h>
#include <linux/dmi.h>
#include <linux/gpio/consumer.h>
#include <linux/gpio/machine.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/jack.h>
#include <linux/input.h>
#include <sound/soc-acpi.h>

#include "../../codecs/hdac_hdmi.h"
#include "hda_dsp_common.h"

#define DXname "DXboard"
#define DX(fmt, ...) pr_debug(DXname " %s: " fmt, __func__, ##__VA_ARGS__)
#define DX1          pr_debug(DXname " %s: >>>\n", __func__)
#define DX0          pr_debug(DXname " %s:  < \n", __func__)
	//dev_dbg(&pdev->dev,"> %s L%d\n",__func__,__LINE__);
	//print_hex_dump_debug;
	//print_hex_dump_bytes;

#define DXGPIO 1

static struct snd_soc_jack hdmi_jack[3];

struct hdmi_pcm {
	struct list_head head;
	struct snd_soc_dai *codec_dai;
//	struct snd_soc_jack hdmi_jack;
	int device;
};


struct sof_card_private {
	struct gpio_desc *gpio_pa;
	struct snd_soc_jack jack;

	bool speaker_en;
	//struct gpio_desc *gpio_48;
	int sample_rate;

	struct list_head hdmi_pcm_list;
//	bool common_hdmi_codec_drv;
};

//----------------------------------------------------------------------------------------------

static int speaker_power_event(struct snd_soc_dapm_widget *w,
	struct snd_kcontrol *kcontrol, int event)
{
	struct snd_soc_card *card = w->dapm->card;
	struct sof_card_private *priv = snd_soc_card_get_drvdata(card);

	if (SND_SOC_DAPM_EVENT_ON(event))
		priv->speaker_en = false;
	else
		priv->speaker_en = true;

	DX("event=%d speaker_en=%d %s\n", event, priv->speaker_en, priv->speaker_en ? "MUTE":"SOUND" );

	gpiod_set_value_cansleep(priv->gpio_pa, priv->speaker_en);
	return 0;
}
//----------------------------------------------------------------------------------------------

static const struct snd_soc_dapm_widget byt_cht_es8316_widgets[] = {
	SND_SOC_DAPM_SPK("Speaker", NULL),
	SND_SOC_DAPM_HP("Headphone", NULL),
	SND_SOC_DAPM_MIC("Headset Mic", NULL),
	SND_SOC_DAPM_MIC("Internal Mic", NULL),

	SND_SOC_DAPM_SUPPLY("Speaker Power", SND_SOC_NOPM, 0, 0,
			    speaker_power_event,
			    SND_SOC_DAPM_PRE_PMD | SND_SOC_DAPM_POST_PMU),
};

static const struct snd_soc_dapm_route byt_cht_es8316_audio_map[] = {
	{"Headphone", NULL, "HPOL"},
	{"Headphone", NULL, "HPOR"},

	/*
	 * There is no separate speaker output instead the speakers are muxed to
	 * the HP outputs. The mux is controlled by the "Speaker Power" supply.
	 */
	{"Speaker", NULL, "HPOL"},
	{"Speaker", NULL, "HPOR"},
	{"Speaker", NULL, "Speaker Power"},

//	{"DMic", NULL, "SoC DMIC"}, //dx

//	{"HDMI1", NULL, "hif1-0 Output"},

//	{ "hifi1", NULL, "iDisp1 Tx"},
//	{ "iDisp1 Tx", NULL, "iDisp1_out"},
};

static const struct snd_soc_dapm_route byt_cht_es8316_intmic_in1_map[] = {
	{"MIC1", NULL, "Internal Mic"},
	{"MIC2", NULL, "Headset Mic"},
};

static const struct snd_soc_dapm_route byt_cht_es8316_intmic_in2_map[] = {
	{"MIC2", NULL, "Internal Mic"},
	{"MIC1", NULL, "Headset Mic"},
};

static const struct snd_kcontrol_new byt_cht_es8316_controls[] = {
	SOC_DAPM_PIN_SWITCH("Speaker"),
	SOC_DAPM_PIN_SWITCH("Headphone"),
	SOC_DAPM_PIN_SWITCH("Headset Mic"),
	SOC_DAPM_PIN_SWITCH("Internal Mic"),
};

static struct snd_soc_jack_pin byt_cht_es8316_jack_pins[] = {
	{
		.pin	= "Headphone",
		.mask	= SND_JACK_HEADPHONE,
	},
	{
		.pin	= "Headset Mic",
		.mask	= SND_JACK_MICROPHONE,
	},
};
//----------------------------------------------------------------------------------------------

static int byt_cht_es8316_init(struct snd_soc_pcm_runtime *runtime)
{
//	struct snd_soc_component *codec = asoc_rtd_to_codec(runtime, 0)->component;  // 5.8
	struct snd_soc_component *codec = runtime->codec_dai->component;  // 5.4

	struct snd_soc_card *card = runtime->card;
	struct sof_card_private *priv = snd_soc_card_get_drvdata(card);
	const struct snd_soc_dapm_route *custom_map;

	int num_routes;
	struct snd_soc_jack *jack;
	int ret;

	DX1;

#if 1
	card->dapm.idle_bias_off = true;
	custom_map = byt_cht_es8316_intmic_in1_map;
	num_routes = ARRAY_SIZE(byt_cht_es8316_intmic_in1_map);
#endif
	
	ret = snd_soc_dapm_add_routes(&card->dapm, custom_map, num_routes);
	if (ret) return ret;

	ret = snd_soc_card_jack_new(card, "Headset",
				    SND_JACK_HEADSET | SND_JACK_BTN_0,
				    &priv->jack, byt_cht_es8316_jack_pins,
				    ARRAY_SIZE(byt_cht_es8316_jack_pins));
	if (ret) {
		dev_err(card->dev, "jack creation failed %d\n", ret);
		return ret;
	}

	snd_jack_set_key(priv->jack.jack, SND_JACK_BTN_0, KEY_PLAYPAUSE);
	snd_soc_component_set_jack(codec, &priv->jack, NULL);

	DX0;
	return 0;
}

static void byt_cht_es8316_exit(struct snd_soc_pcm_runtime *rtd)
{
//	struct snd_soc_component *component = asoc_rtd_to_codec(rtd, 0)->component;  // 5.8
	struct snd_soc_component *component = rtd->codec_dai->component; // 5.4

	DX1;
	snd_soc_component_set_jack(component, NULL, NULL);
	DX0;
}

//----------------------------------------------------------------------------------------------

static int sof_es8336_hw_params(struct snd_pcm_substream *substream,
				struct snd_pcm_hw_params *params)
{
//	struct snd_soc_pcm_runtime *rtd = asoc_substream_to_rtd(substream);     // 5.11
	struct snd_soc_pcm_runtime *rtd = substream->private_data;              // 5.8

	struct sof_card_private *ctx = snd_soc_card_get_drvdata(rtd->card);

//	struct snd_soc_dai *codec_dai = asoc_rtd_to_codec(rtd, 0); // 5.8
	struct snd_soc_dai *codec_dai = rtd->codec_dai;  // 5.4

	struct snd_soc_component *codec = codec_dai->component;

	const int sysclk = 19200000; /* This is fixed on this board */
	int samplerate;
	long mclk_freq=19200000;
	int mclk_div;
	int sampling_freq;
	//bool clk_44;
	int ret;

	
	samplerate = params_rate(params);
	samplerate = 48000;

	DX1;

	if (samplerate == ctx->sample_rate)
		return 0;

	ctx->sample_rate = 0;
/*
 	if (!(IS_ERR_OR_NULL(ctx->gpio_pa))) {
 		printk("%s, set 'gpio-pa' pin to false\n", __func__);
 		gpiod_set_value_cansleep(ctx->gpio_pa, false);
 	}
*/
	ret = snd_soc_dai_set_sysclk(codec_dai, 1, sysclk, SND_SOC_CLOCK_OUT);
	if (ret < 0) {
		printk("%s, Failed to set WM8804 SYSCLK: %d\n", __func__, ret);
		return ret;
	}

/*	ret = snd_soc_dai_set_tdm_slot(codec_dai, 0x3, 0x3, 2, 24);
	if (ret < 0) {
		printk("%d\n",params_width(params));
		printk("%s, set TDM slot err:%d\n", __func__, ret);
		//return ret;
	}
*/
	/* set sampling frequency status bits */

	ctx->sample_rate = samplerate;

	DX0;
	return 0;
}
//----------------------------------------------------------------------------------------------

static int ssp_fixup(struct snd_soc_pcm_runtime *rtd,
			struct snd_pcm_hw_params *params)
{
	struct snd_interval *rate = hw_param_interval(params,
			SNDRV_PCM_HW_PARAM_RATE);
	struct snd_interval *chan = hw_param_interval(params,
			SNDRV_PCM_HW_PARAM_CHANNELS);
	struct snd_mask *fmt = hw_param_mask(params, SNDRV_PCM_HW_PARAM_FORMAT);

	DX1;

	/* The ADSP will convert the FE rate to 48k, stereo */
	rate->min = rate->max = 48000;
	chan->min = chan->max = 2;

	/* set SSP to 24 bit */
	snd_mask_none(fmt);
	snd_mask_set_format(fmt, SNDRV_PCM_FORMAT_S24_LE);

	DX0;
	return 0;
}

/* machine stream operations */
static struct snd_soc_ops sof_es8336_ops = {
	.hw_params = sof_es8336_hw_params,
};
//----------------------------------------------------------------------------------------------

static struct snd_pcm *hda_dsp_hdmi_pcm_handle(struct snd_soc_card *card,
					       int hdmi_idx)
{
	struct snd_soc_pcm_runtime *rtd;
	struct snd_pcm *spcm;
	int i = 0;

	for_each_card_rtds(card, rtd) {
		spcm = rtd->pcm ?
			rtd->pcm->streams[SNDRV_PCM_STREAM_PLAYBACK].pcm : NULL;
		if (spcm && strstr(spcm->id, "HDMI")) {
			if (i == hdmi_idx)
				return rtd->pcm;
			++i;
		}
	}
	return NULL;
}
int hda_dsp_hdmi_build_controls(struct snd_soc_card *card,
				struct snd_soc_component *comp)
{
	struct hdac_hda_priv *hda_pvt;
	struct hda_codec *hcodec;
	struct snd_pcm *spcm;
	struct hda_pcm *hpcm;
	int err = 0, i = 0;

	if (!comp)
		return -EINVAL;

	DX1;

	hda_pvt = snd_soc_component_get_drvdata(comp);
	hcodec = &hda_pvt->codec;

	list_for_each_entry(hpcm, &hcodec->pcm_list_head, list) {
		spcm = hda_dsp_hdmi_pcm_handle(card, i);
		if (spcm) {
			hpcm->pcm = spcm;
			hpcm->device = spcm->device;
			dev_dbg(card->dev,
				"%s: mapping HDMI converter %d to PCM %d (%p)\n",
				__func__, i, hpcm->device, spcm);
		} else {
			hpcm->pcm = NULL;
			hpcm->device = SNDRV_PCM_INVALID_DEVICE;
			dev_warn(card->dev,
				 "%s: no PCM in topology for HDMI converter %d\n\n",
				 __func__, i);
		}
		i++;
	}
	snd_hdac_display_power(hcodec->core.bus, HDA_CODEC_IDX_CONTROLLER, true);

	err = snd_hda_codec_build_controls(hcodec);
	if (err < 0)
		dev_err(card->dev, "unable to create controls %d\n", err);

	snd_hdac_display_power(hcodec->core.bus, HDA_CODEC_IDX_CONTROLLER, false);

	DX0;
	return err;
}
#define NAME_SIZE	32
static int card_late_probe(struct snd_soc_card *card)
{
	struct sof_card_private *ctx = snd_soc_card_get_drvdata(card);
	struct snd_soc_component *component = NULL;
	char jack_name[NAME_SIZE];
	struct hdmi_pcm *pcm;
	int err = 0;
	int i = 0;

	DX1;

	pcm = list_first_entry(&ctx->hdmi_pcm_list, struct hdmi_pcm, head);

	component = pcm->codec_dai->component;

	DX( "pcm: device=%d component=%s", pcm->device, component->name );

	if (!component) return -EINVAL;

	err = hda_dsp_hdmi_build_controls(card, component);

	DX0;
	return err;
}

static int hdmi_init(struct snd_soc_pcm_runtime *rtd)
{
	struct sof_card_private *ctx = snd_soc_card_get_drvdata(rtd->card);
//	struct snd_soc_dai *dai = asoc_rtd_to_codec(rtd, 0); // 5.8
	struct snd_soc_dai *dai = rtd->codec_dai; // 5.4
	struct hdmi_pcm *pcm;

	pcm = devm_kzalloc(rtd->card->dev, sizeof(*pcm), GFP_KERNEL);
	if (!pcm)
		return -ENOMEM;

	pcm->device = 0 + dai->id;
	pcm->codec_dai = dai;

	DX("hdmi_init: dai id=%d name=%s\n",dai->id, dai->name);

	list_add_tail(&pcm->head, &ctx->hdmi_pcm_list);

	return 0;
}
//----------------------------------------------------------------------------------------------

SND_SOC_DAILINK_DEF(ssp0_pin,
	DAILINK_COMP_ARRAY(COMP_CPU("SSP0 Pin")));
SND_SOC_DAILINK_DEF(ssp0_codec,
	DAILINK_COMP_ARRAY(COMP_CODEC("i2c-ESSX8336:00", "ES8316 HiFi")));

//SND_SOC_DAILINK_DEF(dmic_pin,
//	DAILINK_COMP_ARRAY(COMP_CPU("DMIC01 Pin")));
//SND_SOC_DAILINK_DEF(dmic16k_pin,
//	DAILINK_COMP_ARRAY(COMP_CPU("DMIC16k Pin")));
//SND_SOC_DAILINK_DEF(dmic_codec,
//	DAILINK_COMP_ARRAY(COMP_CODEC("dmic-codec", "dmic-hifi")));

SND_SOC_DAILINK_DEF(idisp1_pin,
	DAILINK_COMP_ARRAY(COMP_CPU("iDisp1 Pin")));
SND_SOC_DAILINK_DEF(idisp1_codec,
	DAILINK_COMP_ARRAY(COMP_CODEC("ehdaudio0D2", "intel-hdmi-hifi1")));

SND_SOC_DAILINK_DEF(idisp2_pin,
	DAILINK_COMP_ARRAY(COMP_CPU("iDisp2 Pin")));
SND_SOC_DAILINK_DEF(idisp2_codec,
	DAILINK_COMP_ARRAY(COMP_CODEC("ehdaudio0D2", "intel-hdmi-hifi2")));

SND_SOC_DAILINK_DEF(idisp3_pin,
	DAILINK_COMP_ARRAY(COMP_CPU("iDisp3 Pin")));
SND_SOC_DAILINK_DEF(idisp3_codec,
	DAILINK_COMP_ARRAY(COMP_CODEC("ehdaudio0D2", "intel-hdmi-hifi3")));

SND_SOC_DAILINK_DEF(platform,
	DAILINK_COMP_ARRAY(COMP_PLATFORM("0000:00:1f.3")));

static struct snd_soc_dai_link dx_dailink[] = {
	/* back ends */
	{
		.name = "SSP0-Codec",
		.id = 0,
		.no_pcm = 1,
		.nonatomic = true,
		.dpcm_playback = 1,
		.dpcm_capture = 1,
		.init = byt_cht_es8316_init,
		//.exit = byt_cht_es8316_exit,
		.ops = &sof_es8336_ops,
		.dai_fmt = SND_SOC_DAIFMT_I2S | SND_SOC_DAIFMT_NB_NF | SND_SOC_DAIFMT_CBS_CFS,
		.be_hw_params_fixup = ssp_fixup,
		SND_SOC_DAILINK_REG(ssp0_pin, ssp0_codec, platform),
	},
/*	{
		.name = "dmic01",
		.id = 1,
		.ignore_suspend = 1,
		.dpcm_capture = 1,
		.no_pcm = 1,
		SND_SOC_DAILINK_REG(dmic_pin, dmic_codec, platform),
	},
	{
		.name = "dmic16k",
		.id = 2,
		.ignore_suspend = 1,
		.dpcm_capture = 1,
		.no_pcm = 1,
		SND_SOC_DAILINK_REG(dmic16k_pin, dmic_codec, platform),
	},
*/	{
		.name = "iDisp1",
		.id = 1,
		.init = hdmi_init,
		.dpcm_playback = 1,
		.no_pcm = 1,
		SND_SOC_DAILINK_REG(idisp1_pin, idisp1_codec, platform),
	},
	{
		.name = "iDisp2",
		.id = 2,
		.init = hdmi_init,
		.dpcm_playback = 1,
		.no_pcm = 1,
		SND_SOC_DAILINK_REG(idisp2_pin, idisp2_codec, platform),
	},
	{
		.name = "iDisp3",
		.id = 3,
		.init = hdmi_init,
		.dpcm_playback = 1,
		.no_pcm = 1,
		SND_SOC_DAILINK_REG(idisp3_pin, idisp3_codec, platform),
	},
};
//----------------------------------------------------------------------------------------------

/* SoC card */
static struct snd_soc_card sof_es8336_card = {
	.name = "essx8336", /* sof- prefix added automatically */
	.owner = THIS_MODULE,
	.dapm_widgets = byt_cht_es8316_widgets,
	.num_dapm_widgets = ARRAY_SIZE(byt_cht_es8316_widgets),
	.dapm_routes = byt_cht_es8316_audio_map,
	.num_dapm_routes = ARRAY_SIZE(byt_cht_es8316_audio_map),
	.controls = byt_cht_es8316_controls,
	.num_controls = ARRAY_SIZE(byt_cht_es8316_controls),
	.fully_routed = true,
	.dai_link = dx_dailink,
	.num_links = ARRAY_SIZE(dx_dailink),
	.late_probe = card_late_probe,
};
//----------------------------------------------------------------------------------------------

static struct gpiod_lookup_table up2_gpios_table = {
	/* .dev_id is set during probe */
	.table = {
		GPIO_LOOKUP("INT34BB:00", 264, "PA_ENABLE", GPIO_ACTIVE_LOW),
		{ },
	},
};
//----------------------------------------------------------------------------------------------

static int sof_es8336_probe(struct platform_device *pdev)
{
	struct snd_soc_card *card;
	struct snd_soc_acpi_mach *mach= pdev->dev.platform_data;
	struct sof_card_private *ctx;
	struct acpi_device *adev;
	int dai_index = 0;
	int ret;
	int i;
	struct snd_soc_dai_link *dai_links;

	DX( "----------------------------------------------------------------->" );

	ctx = devm_kzalloc(&pdev->dev, sizeof(*ctx), GFP_KERNEL);
	if (!ctx)
		return -ENOMEM;

	INIT_LIST_HEAD(&ctx->hdmi_pcm_list);

	mach = pdev->dev.platform_data;

	card = &sof_es8336_card;
	card->dev = &pdev->dev;


	if( DXGPIO )
	{
		DX("get gpio resource\n");

		up2_gpios_table.dev_id = dev_name(&pdev->dev);

		gpiod_add_lookup_table(&up2_gpios_table);

		ctx->gpio_pa = devm_gpiod_get(&pdev->dev, "PA_ENABLE", GPIOD_OUT_HIGH); // 1-Mute, 0-Play

		if (IS_ERR(ctx->gpio_pa)) {
			ret = PTR_ERR(ctx->gpio_pa);
			printk("could not get PA_ENABLE: %d\n", ret);
			return ret;
		}

		//printk( "DXboard: gpio %X\n", ctx->gpio_pa );
	}

	ret = snd_soc_fixup_dai_links_platform_name(&sof_es8336_card, mach->mach_params.platform);
	if (ret) return ret;

	snd_soc_card_set_drvdata(card, ctx);

	ret = devm_snd_soc_register_card(&pdev->dev, card);
	
	DX( "--------------------------------------------------------------done" );
	return ret;
}
//----------------------------------------------------------------------------------------------

static int sof_es8336_remove(struct platform_device *pdev)
{
	DX1;
	if( DXGPIO )
		gpiod_remove_lookup_table(&up2_gpios_table);
	DX0;
	return 0;
}
//----------------------------------------------------------------------------------------------

static struct platform_driver sof_es8336_driver = {
	.driver = {
		.name = "sof-essx8336",
		.pm = &snd_soc_pm_ops,
	},
	.probe = sof_es8336_probe,
	.remove = sof_es8336_remove,
};
module_platform_driver(sof_es8336_driver);

MODULE_DESCRIPTION("ASoC Intel(R) SOF + ES8336 Machine driver");
MODULE_AUTHOR("zhuning@everest-semi.com");
MODULE_LICENSE("GPL v2");
MODULE_ALIAS("platform:sof-essx8336");
